include('util.js');             // do_dehydra_dump()

const BAD_NONNULL_ATTR = '__attribute__((user("nonnull"))) specified on non-typedef, or non-pointer typedef';
const BAD_NONNULL_CASTER = '__attribute__((user("nonnull cast"))) specified on non-function, or for non-"nonnull" type, or for function with invalid type signature';
const EXPORTED_NONNULL_PARAM = "exported functions cannot have `nonnull' params";
const ASSN_NULL_TO_NONNULL = "assigning NULL to a `nonnull' type";
const ASSN_NULLABLE_TYPE_TO_NONNULL = "assigning value of nullable type to `nonnull' type";
const PASSING_NULLABLE_TYPE_AS_NONNULL = "passing argument of nullable type to parameter of `nonnull' type";
const RETURNING_NULLABLE_AS_NONNULL = "returning value of nullable type from function with `nonnull` return type";


// TODO customize me for the JS API
function isExportedDecl(decl) {
    return decl.isExtern || !decl.isStatic;
};

// TODO maybe customize based on weird JS stuff
function isNullableToNonnullableCast(fromType, toType) {
    return isNonnullType(toType)
        && !isSameType(fromType, toType);
}

//-----------------------------------------------------------------------------

// TODO decl_map for dehydra

// nonnullTypes: String -> NonnullTypeSpec
let nonnullTypes = {
};

// castFunSet: set(String)
let castFunSet = {
};

function getattribute(decl, attrName) {
    for each (let attr in decl.attributes)
        if (attrName === attr.name)
            return attr.value[0];
    return undefined;
}

function fndeclName(fndecl) {
    return fndecl.name.slice(0, fndecl.name.indexOf('('));
}

function isSameType(t1, t2) {
    return t1.name === t2.name;
}

function isNonnullType(t) {
    return t.name in nonnullTypes;
}

function fnTypeHasNonnullParam(ftype) {
    for each (let param in ftype.parameters) {
        if (isNonnullType(param))
            return true;
    }
    return false;
}

function isConstant(node) {
    // TODO can do better than this?
    return node.hasOwnProperty('value');
}

//-----------------------------------------------------------------------------
// In the following `tcheck' functions:
//   - each returns the type it infers
//   - `cx' means the context in which the statement/expr appears
//     (usually a fndecl)
//   - `loc' is the location to blame if an error is found
//

function tcheckNode(cx, loc, node) {
    if (node.isReturn)
        // don't return here: isReturn does not exclude other node types
        tcheckReturn(cx, loc, node);
    if (node.assign)
        return tcheckAssn(cx, loc, node, node.assign[0]);
    if (node.isFcall)
        return tcheckFuncall(cx, loc,
                             node, node.arguments, node.type.parameters);
    //if (node.isDecl)  TODO hmm.  need to handle this?
    return node.type;
}

function tcheckAssn(cx, loc, lhs, rhs) {
    let lhsType = lhs.type;
    let rhsType = tcheckNode(cx, loc, rhs);

    if (!isNonnullType(lhsType))
        return lhsType;

    // Rule 2: the constant '0' cannot be assigned to a "nonnull" type
    if (isConstant(rhs) && '0u' === rhs.value)
        error(ASSN_NULL_TO_NONNULL, loc);
    // Rule 3: only a decl of the "nonnull" pointer type can
    //   be assigned to a "nonnull"-typed variable
    else if (!isNullableToNonnullableCast(rhsType, lhsType))
        error(ASSN_NULLABLE_TYPE_TO_NONNULL, loc);

    return lhsType;
}

function tcheckFuncall(cx, loc, callable, args, params) {
    let retType = callable.type.type;

    for (let i = 0; i < args.length; ++i) {
        // Rule 4: only a decl of the "nonnull" pointer type can
        //   be assigned to a "nonnull"-typed function parameter

        let argType = tcheckNode(cx, loc, args[i]);
        if (undefined === params[i]) {
            warning("insufficient type information for param "+ i, loc);
            continue;           // TODO warn?
        }
        let paramType = params[i];

        if (isNullableToNonnullableCast(argType, paramType)) {


            error(PASSING_NULLABLE_TYPE_AS_NONNULL, loc);


            if (0) {
                print ('---'+ callable.name);
                print ('  ARG');
                do_dehydra_dump(argType, 4, 1);
                print ('  PARAM');
                do_dehydra_dump(paramType, 4, 1);
            }
        }

    }
    return retType;
}

function tcheckReturn(cx, loc, ret) {
    let retExprType = ret.type.type;
    let cxRetType = cx.type.type;

    // Rule 5: only a decl of the "nonnull" pointer type can
    //   be returned from a function with "nonnull" return type
    if (isNullableToNonnullableCast(retExprType, cxRetType))
        error(RETURNING_NULLABLE_AS_NONNULL, loc);

    return ret.type;
}

//-----------------------------------------------------------------------------
// "Main"

function process_function(fndecl, body) {
    let fnname = fndeclName(fndecl);

    // we don't analyze nonnull `cast' functions
    if (fnname in castFunSet) {
//        print('(  not checking: '+ fnname);
        return;
    }

//    print('----------------- CHECKING:'+ fnname);

    for each (let block in body)
        for each (let stmt in block.statements)
            tcheckNode(fndecl, block.loc, stmt);

    // XXX nicety: it should be an error to use an uninitialized "nonnull"
    // XXX decl.  Should not be just a warning.
}

function process_decl(decl) {
    let userattr = getattribute(decl, 'user');

    if (undefined !== userattr) {
        switch (userattr) {
        case 'nonnull':
            if (!(decl.type
                  && decl.type.typedef
                  && decl.type.typedef.isPointer)) {
                error(BAD_NONNULL_ATTR);
                return;
            }
            // found an __attribute__((user("nonnull"))) decl.  next we need a
            // non-null casting function, with attribute 'nonnull cast'.

            // TODO decl_map for dehydra
            nonnullTypes[decl.name] = {
                'base': decl.type.typedef,
                'typedef': decl.type,
                'cast': undefined
            };
            break;

        case 'nonnull cast':
            if (!(decl.isFunction
                  && 1 === decl.parameters.length))
                error(BAD_NONNULL_CASTER);

            let ftype = decl.type;
            let nonnullType = nonnullTypes[ftype.type.name];
            if (!(undefined !== nonnullType
                  && nonnullType.base == decl.parameters[0].type))
                error(BAD_NONNULL_CASTER);

            nonnullType.cast = decl;
            castFunSet[fndeclName(decl)] = true;
            break;

        default:
            // ???
            break;
        }

        return;
    }

    // Rule 1: a "nonnull" type cannot be a parameter to an "extern" function
    if (decl.isFunction
        && isExportedDecl(decl)
        && fnTypeHasNonnullParam(decl.type)) {
        error(EXPORTED_NONNULL_PARAM, decl.loc);
        return;
    }
}
